#!/usr/bin/env python

__author__ = "Patrick Godwin (patrick.godwin@ligo.org)"
__description__ = "tools to serve data and dynamic html pages"

#-------------------------------------------------
### imports

import json
import os
import sys
import time

import bottle
import numpy

from lal import gpstime

from . import aggregator
from . import utils

#-------------------------------------------------
### constants + env variables

JSON_HEADER = {
    'Content-type': 'application/json',
    'Cache-Control': 'max-age=10',
}

NUM_JOBS = 350

#-------------------------------------------------
### bottle apps/routing

app = bottle.Bottle()

@app.route("/query")
def mock_query():
    ### parse query
    try:
        database = bottle.request.query['db']
        query = bottle.request.query['q'].strip()
        epoch = bottle.request.query['epoch']
        start, end, measurements, aggregate, tags, filters, latest = parse_query(query)
    except ValueError as e:
        return bottle.HTTPResponse(status=400, headers=JSON_HEADER, body=json.dumps({'error': str(e), 'query': repr(query)}))

    ### generate fake data
    start = utils.unix_to_gps(start)
    end = utils.unix_to_gps(end)
    series = generate_timeseries(start, end, measurements, aggregate, tags=tags, filters=filters, epoch=epoch, latest=latest)
    response = {'results': [{'statement_id':0, 'series':series}]}

    ### return data
    return bottle.HTTPResponse(status=200, headers=JSON_HEADER, body=json.dumps(response))


#-------------------------------------------------
### functions

def parse_query(query_str):
    ### find if limit is in query, handle case separately
    query = query_str.strip().strip("'").split()
    if 'ORDER BY time DESC' in query_str and 'LIMIT' in query_str:
        latest = int(query[-1])
        query = query[:len(query)-6]
        end = aggregator.now()
        start = end - 1000
        start, end = utils.gps_to_unix(start), utils.gps_to_unix(end)
    else:
        latest = 0

    ### basic query statement validations
    for statement in ('SELECT', 'FROM', 'WHERE'):
        if statement not in query_str:
            return ValueError('missing {}'.format(statement))

    if not latest:
        for comparison in ('>=', '<='):
            if 'time {}'.format(comparison) not in query_str:
                return ValueError('missing time {}'.format(comparison))

    ### parse select and validate
    select_statement, field_tags = query[0], query[1]
    if select_statement != 'SELECT':
        return ValueError('SELECT not in proper location')
    field_tags = field_tags.split(',')
    field, tags = field_tags[0], field_tags[1:]

    ### parse from and validate
    from_statement, measurements = query[2], query[3]
    if from_statement != 'FROM':
        return ValueError('FROM not in proper location')
    measurements = measurements.split(',')

    ### parse where and validate
    where_statement, conditions = query[4], query[5:]
    if where_statement != 'WHERE':
        return ValueError('WHERE not in proper location')
    filters = []
    conditions = ' '.join(conditions).split('AND')
    for condition in conditions:
        column, symbol, val = condition.strip().split()
        val = val.strip("'")
        if column == 'time':
            if symbol == '>=':
                start = int(val)
            elif symbol == '<=':
                end = int(val)
            else:
                return ValueError('not a supported time comparison operator')
        else:
            if symbol != '=':
                return ValueError('tag only supports = operator')
            if column == 'aggregate':
                aggregate = val
            else:
                filters.append((column, val))

    return start, end, measurements, aggregate, tags, filters, latest


def generate_timeseries(start, end, measurements, aggregate, tags=None, filters=None, epoch='ns', latest=0):
    ### format filters
    ### FIXME: doesn't do anything for now
    columns = []
    for filter_ in filters:
        tag, val = filter_
        columns.append(tag)

    ### build up timeseries
    all_series = []
    for measurement in measurements:
        _, dt, series_name = measurement.split('.')
        dt = dt.strip('"')[:-1]
        series_name = series_name.strip('"')
        series = []
        if tags and tags[0].strip('"') == 'job': ### FIXME: assume 1 tag max for now
            for job_id in range(NUM_JOBS):
                if latest:
                    for t in numpy.random.randint(start, high=end, size=latest):
                        if epoch == 'ns':
                            datetime = utils.gps_to_unix(t)
                        elif epoch == 'datetime':
                            datetime = gpstime.gps_to_str(t, "%Y-%m-%dT%H:%M:%S.%fZ")
                        series.append([datetime, numpy.random.normal(), job_id])
                else:
                    for t in numpy.arange(start, end, int(dt)):
                        if epoch == 'ns':
                            datetime = utils.gps_to_unix(t)
                        elif epoch == 'datetime':
                            datetime = gpstime.gps_to_str(t, "%Y-%m-%dT%H:%M:%S.%fZ")
                        series.append([datetime, numpy.random.normal(), job_id])
        else:
            if latest:
                for t in numpy.random.randint(start, high=end, size=latest):
                    if epoch == 'ns':
                        datetime = utils.gps_to_unix(t)
                    elif epoch == 'datetime':
                        datetime = gpstime.gps_to_str(t, "%Y-%m-%dT%H:%M:%S.%fZ")
                    series.append([datetime, numpy.random.normal()])
            else:
                for t in numpy.arange(start, end, int(dt)):
                    if epoch == 'ns':
                        datetime = utils.gps_to_unix(t)
                    elif epoch == 'datetime':
                        datetime = gpstime.gps_to_str(t, "%Y-%m-%dT%H:%M:%S.%fZ")
                    series.append([datetime, numpy.random.normal()])
        all_series.append({
            'name': series_name,
            'columns': ['time', 'data'],
            'values': series,
        })

    return all_series


def _add_parser_args(parser):
    parser.add_argument('-b', '--backend', default='wsgiref',
                        help="chooses server backend. options: [cgi|wsgiref]. wsgiref starts a local server for development, default = wsgiref.")
    parser.add_argument('-p', '--port', type=int, default=8086,
                        help="select port to serve content on server. default = 8086")


#-------------------------------------------------
### main

def main(args=None):
    """Mocks an InfluxDB database

    """
    if not args:
        parser = argparse.ArgumentParser()
        _parser_add_arguments(parser)
        args = parser.parse_args()

    port = args.port
    server_backend = args.backend

    bottle.run(app, server=server_backend, port=port, debug=True)
